/*
 * libEBjabber
 *
 * Copyright (C) 2000, Alex Wheeler <awheeler@speakeasy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	This code borrowed heavily from the jabber client jewel
 *	(http://jewel.sourceforge.net/)
 */

#include <stdio.h>
#include "libEBjabber.h"

/******************************************************
 * Support routines for mulitple simultaneous accounts
 *****************************************************/
JABBER_Conn JabberConnection;

JABBER_Conn *JCnewConn(void)
{
	/* Create a new connection struct*/
	//JabberConnection=(JABBER_Conn *)calloc(1, sizeof(JABBER_Conn));
	ZeroMemory(&JabberConnection, sizeof(JABBER_Conn));
	return(&JabberConnection);
}

/*JABBER_Conn *JCfindConn(jconn conn)
{
	JABBER_Conn *current=JabberConnection;

	while(current)
	{
		if(conn == current->conn)
			return(current);
//		current=current->next;
	}
	return(NULL);
}*/

char *JCgetServerName(JABBER_Conn * JConn)
{
	if(!JConn || !JConn->conn || !JConn->conn->user)
		return(NULL);
	return(JConn->conn->user->server);
}


char **JCgetJIDList(void)
{
	if(GetJabberConnState() == JCONN_STATE_OFF) return NULL;
	JABBER_Conn *current=&JabberConnection;
	char **list=NULL;
	int count=0;

	/* Leave room for the last null entry */
	list=(char **)realloc(list, sizeof(char *)*(count+2));
	list[count]=strdup(current->jid);
	count++;
	if(list) list[count]=NULL;
	return(list);
}

void JCfreeServerList(char **list)
{
	int count=0;
	while(list[count])
	{
		free(list[count++]);
	}
	free(list);
}

/****************************************
 * End jabber multiple connection routines
 ****************************************/

/* Functions called from jabber.c file */
JABBER_Conn *JABBER_Login(char *handle, char *passwd, BOOL newAccount) 
{
	/* At this point, we don't care about host and port */
	// int  lID;
	char jid[256+1];
	char buff[4096];
	JABBER_Conn *JConn=NULL;
	// xmlnode x;

	// printf( "%s %s %i\n", handle, host, port);
	/* Make sure the name is correct, support all formats
	 * handle	become handle@host/Freewebfone
	 * handle@foo	becomes handle@foo/Freewebfone
	 * handle@foo/bar is not changed
	 */

	if(!strchr(handle, '@'))
	{
		/* A host must be specified in the config file */
//		if(!host) {
			JABBERError("Error in jabber-login");
			return(NULL);
//		}
	}
	else strncpy(jid, handle, 256);
	// printf( "jid: %s\n", jid);
	JConn=JCnewConn();
	strncpy(JConn->jid,jid, LINE_LENGTH);
	JConn->conn = jab_new(jid, passwd);
	if(!JConn->conn)
	{
		snprintf(buff, 4096, "Connection to the jabber server failed!");
		JABBERError(buff);
		JConn = NULL;
		logoutJabber();
		return(NULL);
	}
	jab_packet_handler(JConn->conn, j_on_packet_handler);
	jab_state_handler(JConn->conn, j_on_state_handler);
	jab_start(JConn->conn);
	if(!GetJabberConnState())
	{
		snprintf(buff, 4096, "Connection to the jabber server failed!");
		JABBERError(buff);
		logoutJabber();
		return(NULL);
	}

	// register if it's a new account
	if (newAccount)
		jab_reg(JConn->conn);
	else
		jab_auth(JConn->conn);


    //    	lID = gdk_input_add(JConn->conn->fd, GDK_INPUT_READ,
    //                        jabber_callback_handler, JConn);
    // Create a thread to receive data from the server
    createJabberRcvThread(JConn);

	// JConn->listenerID = lID;
	// printf( "*** ListenerID: %i FD: %i\n", lID, JConn->conn->fd);
	/* FIXME: Is this supposed to be 1? */
	JConn->id = 1;
	JConn->reg_flag = 0;

	return(JConn);
}

int JABBER_Logout()
{
	if(JabberConnection.conn)
	{
		// printf( "JConn->conn exists, closing everything up\n");
		//jab_stop(JConn->conn);
		jab_delete(JabberConnection.conn);
		JabberConnection.conn=NULL;
	}
	//free(JabberConnection);
	//JabberConnection = NULL;

	//Logout complete
	return(0);
}

int JABBER_AuthorizeContact(JABBER_Conn *conn, char *handle) {
	// printf( "%s\n", handle);
	return(0);
}

int JABBER_SendMessage(JABBER_Conn *JConn, char *handle, char *message) {
	xmlnode x;

	if(!JConn) {
		// printf("******Called with NULL JConn for user %s!!!\n", handle);
		return(0);
	}
	// printf( "handle: %s message: %s\n", handle, message);
	// printf( "********* %s -> %s\n", JConn->jid, handle);
	/* We always want to chat.  :) */
	x = jutil_msgnew(TMSG_CHAT, handle, NULL, message);
	jab_send(JConn->conn, x);
	return(0);
}

int JABBER_AddContact(JABBER_Conn *JConn, char *handle) {
	xmlnode x,y,z;
	char *jid=strdup(handle);
	JABBER_Dialog *JD;
	char *buddy_server=NULL;
	char **server_list;

	if(!JConn) {
		buddy_server=strchr(handle, '@');
		if(!buddy_server) {
			/* This happens for tranposrt buddies */
			buddy_server=strchr(handle, '.');
			if(!buddy_server) {
				// printf("Something weird, buddy without an '@' or a '.'");
				return(1);
			}
		}
		buddy_server++;
		JConn=&JabberConnection;
		if(JConn) {
			;// printf("Found matching server for %s\n", handle);
		}
		else {
			server_list=JCgetJIDList();
			if(!server_list)
				return(1);
			JD=(JABBER_Dialog *)calloc(sizeof(JABBER_Dialog), 1);
			JD->heading="Pick an account";
			JD->message=strdup("Unable to automatically determine which account to use for thise buddy:\nplease select the account that can talk to this buddy's server");
			JD->callback=j_on_pick_account;
			JD->requestor=strdup(handle);
			JABBERListDialog(server_list, JD);
			return(0);
		}
	}
	/* For now, we do not understand resources */
	jid=strtok(jid, "/");
	if(!jid)
		jid=strdup(handle);
	// printf( "%s now %s\n", handle, jid);

	/* Ask to be subscribed */
	x=jutil_presnew(JPACKET__SUBSCRIBE, jid, NULL);
	jab_send(JConn->conn, x);

	/* Request roster update */
	x = jutil_iqnew (JPACKET__SET, NS_ROSTER);
	y = xmlnode_get_tag (x, "query");
	z = xmlnode_insert_tag (y, "item");
	xmlnode_put_attrib (z, "jid", jid);
	jab_send (JConn->conn, x);
	return(0);
}

int JABBER_RemoveContact(JABBER_Conn *JConn, char *handle) {
	xmlnode x,y,z;
	char *jid=strdup(handle);

	if(!JConn) {
		// fprintf(stderr, "**********NULL JConn sent to JABBER_RemoveContact!\n");
		return(-1);
	}

	x = jutil_presnew (JPACKET__UNSUBSCRIBE, handle, NULL);
	jab_send (JConn->conn, x);

	x = jutil_iqnew (JPACKET__SET, NS_ROSTER);
	y = xmlnode_get_tag (x, "query");
	z = xmlnode_insert_tag (y, "item");
	xmlnode_put_attrib (z, "jid", jid);
	xmlnode_put_attrib (z, "subscription", "remove");
	jab_send (JConn->conn, x);

	return(0);
}

int JABBER_EndChat(JABBER_Conn *JConn, char *handle) {
	// printf( "Empty function\n");
	return(0);
}



int JABBER_ChangeState(JABBER_Conn *JConn, int state) {
	xmlnode x,y;
	/* Unique away states are possible, but not supported 
	 * yet.  status would hold that value */
	char show[7]="", *status="";

	// printf("(%i)\n", state);
	if(!JConn->conn)
		return(-1);
	x = jutil_presnew (JPACKET__UNKNOWN, NULL, NULL);

	if (state != JABBER_ONLINE) {
		y = xmlnode_insert_tag (x, "show");
		switch (state) {
		case JABBER_AWAY:
			strcpy (show, "away");
			break;
		case JABBER_DND:
			strcpy (show, "dnd");
			break;
		case JABBER_XA:
			strcpy (show, "xa");
			break;
		case JABBER_CHAT:
			strcpy (show, "chat");
			break;
		default:
			strcpy (show, "unknown");
			// printf( "Unknown state: %i suggested\n", state);
		}
		xmlnode_insert_cdata (y, show, -1);
	}
	if (strlen (status) > 0) {
		y = xmlnode_insert_tag (x, "status");
		xmlnode_insert_cdata (y, status, -1);
	}
	// printf( "Setting status to: %s - %s\n", show, status);
	jab_send (JConn->conn, x);
	return(0);
}

/* ----------------------------------------------*/
/* Functions to handle jabber responses */
void j_on_packet_handler(jconn conn, jpacket packet) 
{
	xmlnode x, y, z;
	char *id, *ns, *jid, *sub, *name, *from, *group;
	char *desc, *service, *to, *url;
	char *show, *type, *body, *subj, *ask, *code;
	char buff[65536+1];	/* FIXME: FInd right size */
	int iid, status;
	JABBER_Conn *JConn=NULL;
	JABBER_InstantMessage JIM;
	JABBER_Dialog *JD;
	struct jabber_buddy JB;

	JConn=&JabberConnection;
	jpacket_reset(packet);
	// printf("Packet: %s\n", packet->x->name);
	switch(packet->type) {
	case JPACKET_MESSAGE:
		// printf( "MESSAGE recieved\n");
		from = xmlnode_get_attrib (packet->x, "from");
		type = xmlnode_get_attrib (packet->x, "type");
		/* type can be empty, chat, groupchat, or error */
		x = xmlnode_get_tag (packet->x, "body");
		if (x) body = xmlnode_get_data (x);
		else body = NULL;
		x = xmlnode_get_tag (packet->x, "subject");
		if (x) {
			subj = xmlnode_get_data (x);
			if (body != NULL ) snprintf(buff, 65536, "%s: %s", subj, body);
			else snprintf(buff, 65536, "%s", subj);
		}
		else {
			subj = NULL;
			if (body != NULL ) strncpy_s(buff, 65536, body, 65536);
			else buff[0] = '\0';
		}
		// fprintf (stderr, "from %s, type %s, body: %s, subj %s\n", from, type, body, subj);
		JIM.msg=buff;
		// printf( "JIM.msg: %s\n", JIM.msg);
		/* For now, we do not understand resources */
//		JIM.sender=strtok(from, "/");
//		if(!JIM.sender)
			JIM.sender=from;
		// printf( "JIM.sender: %s\n", JIM.sender);
		// printf( "Rendering message\n");
		JABBERInstantMessage(&JIM);
		break;
	case JPACKET_IQ:
		type = xmlnode_get_attrib(packet->x, "type");
		if(!type) {
			// printf( "JPACKET_IQ: NULL type\n");
			return;
		}
		// printf( "IQ recieved: %s\n", type);
		if(!strcmp(type, "result")) {
			id = xmlnode_get_attrib (packet->x, "id");
			if (id) {
				iid = atoi (id);
				// printf( "iid: %i\n", iid);
				if (iid == JConn->id) {
					if (!JConn->reg_flag) {
						// fprintf (stderr, "requesting roster\n");
						x = jutil_iqnew (JPACKET__GET, NS_ROSTER);
						jab_send (conn, x);
						// fprintf (stderr, "requesting agent list\n");
						x = jutil_iqnew (JPACKET__GET, NS_AGENTS);
						jab_send (conn, x);
						return;
					}
					else {
						JConn->reg_flag = 0;
						JConn->id = atoi (jab_auth (conn));
						return;
					}
				}
			}
/*
			else {
				printf( "No ID!\n");
			}
*/
			x = xmlnode_get_tag (packet->x, "query");
			if (!x) {
				// printf( "No X\n");
				break;
			}
			ns = xmlnode_get_attrib (x, "xmlns");
			// printf( "ns: %s\n", ns);
			if (strcmp (ns, NS_ROSTER) == 0) {
				y = xmlnode_get_tag (x, "item");
				for (;;) {
					jid = xmlnode_get_attrib (y, "jid");
					sub = xmlnode_get_attrib (y, "subscription");
					name = xmlnode_get_attrib (y, "name");
					group = xmlnode_get_attrib (y, "group");
					// printf( "Buddy: %s(%s) %s %s\n", jid, group, sub, name);
					if(jid && sub && name) {
						JB.jid=strtok(jid, "/");
						if(!JB.jid) {
							JB.jid=strdup(jid);
						}
						else {
							JB.jid=strdup(JB.jid);
						}
						JB.name=strdup(name);
						JB.sub=strdup(sub);
						/* State does not matter */
						JB.status=JABBER_OFFLINE;
						JB.JConn=JConn;
						jabberAddBuddy(&JB);
					}
					y = xmlnode_get_nextsibling (y);
					if (!y) break;
				}
			}
			else if(strcmp (ns, NS_AGENTS)==0) {
				y = xmlnode_get_tag (x, "agent");
				for (;;) {
					jid = xmlnode_get_attrib (y, "jid");
					if(jid) {
						name = xmlnode_get_tag_data (y, "name");
						desc = xmlnode_get_tag_data (y, "description");
						service = xmlnode_get_tag_data (y, "service");
						// printf( "Agent[%s]: %s [%s] [%s]\n",
						//	name, jid, desc, service);
						z = jutil_iqnew (JPACKET__GET, NS_AGENT);
						xmlnode_put_attrib(z, "to", jid);
//						xmlnode_put_attrib(z, "id", "1");
						to=xmlnode_get_attrib(z, "to");
						/* printf("to: %s - %s - %s\n", to,
							xmlnode_get_attrib(z, "type"),
							xmlnode_get_attrib(z, "id"));
						*/
						jab_send (conn, z);
					}
					y = xmlnode_get_nextsibling (y);
					if (!y) break;
				}
			}
			else if(strcmp (ns, NS_AGENT)==0) {
				name=xmlnode_get_tag_data(x, "name");
				url =xmlnode_get_tag_data(x, "url");
				// printf("agent: %s - %s\n", name, url);
			}
			jab_send_raw (conn, "<presence/>");
			return;
			
		}
		else
		if(!strcmp(type, "set")) {
			x = xmlnode_get_tag (packet->x, "query");
			ns = xmlnode_get_attrib (x, "xmlns");
			if (strcmp (ns, NS_ROSTER) == 0) {
				y = xmlnode_get_tag (x, "item");
				jid = xmlnode_get_attrib (y, "jid");
				sub = xmlnode_get_attrib (y, "subscription");
				name = xmlnode_get_attrib (y, "name");
				ask = xmlnode_get_attrib (y, "ask");
				// printf( "Need to find a buddy: %s %s %s %s\n", jid, sub, name, ask);
/*				bud = buddy_find (jid);
*/
				if (sub) {
					if (strcmp (sub, "remove") == 0) {
						// printf( "Need to remove a buddy: %s\n", jid);
/*						buddy_remove_from_clist (bud);
						buddy_remove (bud);
*/
						break;
					}
				}
/*
				if (!bud) {
					buddy_add_to_clist (buddy_add (jid, sub, name));
					if (!name && !ask) roster_new ("Edit User", jid, NULL);
				}
				else if (name) {
					free (bud->name);
					bud->name = strdup (name);
					set_name (bud);
				}
*/
			}

		}
		else
		if(!strcmp(type, "error"))
		{
			x = xmlnode_get_tag (packet->x, "error");
			code = xmlnode_get_attrib (x, "code");
			desc = xmlnode_get_tag_data(packet->x, "error");

			// printf("Recieved error: %s\n", desc);
			switch(atoi(code))
			{
			case 302: /* Redirect */
				JABBERError("Jabber Eror 302. Redirect");
				break;
			case 400: /* Bad request */
				JABBERError("Jabber Error 402. Bad request");
				break;
			case 401: /* Unauthorized */
				JABBERError("Jabber Error. Not authorized. Check Login and Password");

/*				JD=(JABBER_Dialog *)calloc(sizeof(JABBER_Dialog), 1);
				JD->heading="Register? You're not authorized";
				JD->message=strdup("Do you want to try to create an account on the jabber server?");
				JD->callback=j_on_create_account;
				JD->JConn=JConn;
				JABBERDialog(JD);
*/				break;
			case 402: /* Payment Required */
				JABBERError("Jabber Error 402. Payment Required");
				break;
			case 403: /* Forbidden */
				JABBERError("Jabber Error 403. Forbidden");
				break;
			case 404: /* Not Found */
				JABBERError("Jabber Error 404. Not Found");
				break;
			case 405: /* Not Allowed */
				JABBERError("Jabber Error 405. Not Allowed");
				break;
			case 406: /* Not Acceptable */
				JABBERError("Jabber Error 406. Not Acceptable");
				break;
			case 407: /* Registration Required */
				JABBERError("Jabber Error 407. Registration Required");
				break;
			case 408: /* Request Timeout */
				JABBERError("Jabber Error 408. Request Timeout");
				break;
			case 409: /* Conflict */
				JABBERError("Jabber Error 409. Conflict");
				break;
			case 500: /* Internal Server Error */
				JABBERError("Jabber Error 500. Internal Server Error");
				break;
			case 501: /* Not Implemented */
				JABBERError("Jabber Error 501. Not Implemented");
				break;
			case 502: /* Remote Server Error */
				JABBERError("Jabber Error 502. Remote Server Error");
				break;
			case 503: /* Service Unavailable */
				JABBERError("Jabber Error 503. Service Unavailable");
				break;
			case 504: /* Remote Server Timeout */
				JABBERError("Jabber Error 504. Remote Server Timeout");
				break;
			default:
				JABBERError("Jabber Unknown error");
				break;
			}
			conn->state = JCONN_STATE_OFF;
			if(conn->on_state) (conn->on_state)(conn, JCONN_STATE_OFF);
			SendStatusInfo(-1);
		}
		break;
	case JPACKET_PRESENCE:
		status = JABBER_ONLINE;
		type = xmlnode_get_attrib (packet->x, "type");
		from = xmlnode_get_attrib (packet->x, "from");
		/* For now, we do not understand resources */
		//JB.jid=strtok(from, "/");
		//if(!JB.jid)
			JB.jid=from;
		// printf( "PRESENCE recieved type: %s from: %s\n", type, from);
		x = xmlnode_get_tag (packet->x, "show");
		if (x) {
			show = xmlnode_get_data (x);
			// printf ("show: %s\n", show);
			if (show) {  // DD
				if (strcmp (show, "away") == 0) status = JABBER_AWAY;
				else if (strcmp (show, "dnd") == 0) status = JABBER_DND;
				else if (strcmp (show, "xa") == 0) status = JABBER_XA;
				else if (strcmp (show, "chat") == 0) status = JABBER_CHAT;
			}
		}
		if (type) {
			if (strcmp (type, "unavailable") == 0) status = JABBER_OFFLINE;
		}
		JB.status=status;
		JB.JConn=JConn;
		JABBERStatusChange(&JB);
		break;
	case JPACKET_S10N:
		// printf( "S10N recieved\n");
		from = xmlnode_get_attrib (packet->x, "from");
		type = xmlnode_get_attrib (packet->x, "type");
		if (type)
		{
			JD=(JABBER_Dialog *)calloc(sizeof(JABBER_Dialog), 1);
			if (strcmp (type, "subscribe") == 0)
			{
				sprintf(buff, "%s wants to subscribe you, would you like to accept it?\n", from);
				JD->message=strdup(buff);
				JD->heading="Subscribe Request";
				JD->callback=j_allow_subscribe;
				JD->requestor=strdup(from);
				JD->JConn=JConn;
				// JABBERDialog(JD);
				//subscribeRequest(JD);
				return;
			}
			else if (strcmp (type, "unsubscribe") == 0)
			{
				x = jutil_presnew (JPACKET__UNSUBSCRIBED, from, NULL);
				jab_send (conn, x);
				// sprintf(buff, "%s unsubscribed you, do you want to unsubscribe them?", from);
				sprintf(buff, "%s unsubscribed you.", from);
				JD->message=strdup(buff);
				JD->heading="Remove User";
				// JD->callback=j_unsubscribe;
				JD->requestor=strdup(from);
				JD->JConn=JConn;
				JABBERDialog(JD);
				unsubscribeRequest(JD);
				return;
			}
		}
		break;
	default:
		// fprintf(stderr, "unrecognized packet: %i recieved\n", packet->type);
		break;
	}
}

void j_on_state_handler(jconn conn, int state)
{
	static int previous_state=JCONN_STATE_OFF;
	JABBER_Conn *JConn=NULL;
	char buff[4096];

	// printf("Entering: new state: %i previous_state: %i\n", state, previous_state);
	JConn=&JabberConnection;
	switch(state) {
	case JCONN_STATE_OFF:
		if(previous_state!=JCONN_STATE_OFF)
		{
			logoutJabber();
			snprintf(buff, 4096, "Disconnected from Jabber server");
			JABBERError(buff);
			Reconnect(5000);
		}
		break;
	case JCONN_STATE_CONNECTED:
		// printf( "JCONN_STATE_CONNECTED\n");
		break;
	case JCONN_STATE_CONNECTING:
		// printf( "JCONN_STATE_CONNECTED\n");
		break;
	case JCONN_STATE_AUTH:
		// printf( "JCONN_STATE_AUTH\n");
		break;
	case JCONN_STATE_ON:
		// printf( "JCONN_STATE_ON\n");
		break;
	default:
		// printf( "UNKNOWN state: %i\n", state);
		break;
	}
	SendStatusInfo(state);
	previous_state=state;
}

void j_on_create_account(void *data) {
	JABBER_Dialog_PTR jd;

	// printf( "Entering, but doing little\n");
	jd=(JABBER_Dialog_PTR)data;
	jd->JConn->reg_flag = 1;
	jd->JConn->id = atoi (jab_reg (jd->JConn->conn));
}

void j_allow_subscribe(void *data) {
	JABBER_Dialog_PTR jd;
	xmlnode x, y, z;

	jd=(JABBER_Dialog_PTR)data;
	// printf( "%s: Entering\n", jd->requestor);
	x=jutil_presnew(JPACKET__SUBSCRIBED, jd->requestor, NULL);
	jab_send(jd->JConn->conn, x);

	/* We want to request their subscription now */
	x=jutil_presnew(JPACKET__SUBSCRIBE, jd->requestor, NULL);
	jab_send(jd->JConn->conn, x);

	/* Request roster update */
	x = jutil_iqnew (JPACKET__SET, NS_ROSTER);
	y = xmlnode_get_tag (x, "query");
	z = xmlnode_insert_tag (y, "item");
	xmlnode_put_attrib (z, "jid", jd->requestor);
	xmlnode_put_attrib (z, "name", jd->requestor);
	jab_send (jd->JConn->conn, x);
}

void j_unsubscribe(void *data) {
	JABBER_Dialog_PTR jd;

	jd=(JABBER_Dialog_PTR)data;
	jabberDelBuddy(jd->requestor);
}

void j_on_pick_account(void *data) {
	JABBER_Dialog_PTR jd;
	JABBER_Conn *JConn=NULL;

	jd=(JABBER_Dialog_PTR)data;
	JConn=&JabberConnection;
	// printf("Found JConn for %s: %p\n", jd->requestor, JConn);
	/* Call it again, this time with a good JConn */
	if(!JConn) {
		// fprintf(stderr, "NULL Jabber connector for buddy, should not happen!\n");
		return;
	}
	JABBER_AddContact(JConn, jd->requestor);
}

